#define  _DEVICE
#include <fstream>
#include <iostream>
#include <algorithm>
#include <cctype>
#include <errno.h>

#include "configs.h"

using namespace std;

bool Cmpi::operator()( const string& x, const string& y) const {
    string::const_iterator p = x.begin();
    string::const_iterator q = y.begin();
    while( p!=x.end() && q!=y.end() && toupper(*p) == toupper(*q) ) { ++p; ++q; }
    if(p==x.end()) return q!=y.end();
    if(q==y.end()) return false;
    return toupper(*p) < toupper(*q);
}

int cfg_file_read( const std::string& fname, Tree& cfgarr, std::vector<std::string>& pr ){
    vector<string> tokens(8);
    int nstr = 0;
    std::string str;
    std::string section, subsection, keyword;
    bool  eos = true, eoss = true, cfgerr = false;
    ifstream file( fname.c_str() );

    if( !file ) throw ErrFat( (ErrorsCode)errno, fname );

    while( getline( file, str ) ){
        if( cfgerr ) throw ErrFat( ERR_CFG_STRUCT, fname );
        nstr++;
        if( !Tool::parse( str.c_str(), tokens ) || tokens.empty()  ) continue;
        keyword = tokens[0];
        transform( keyword.begin(), keyword.end(), keyword.begin(), ptr_fun(::tolower) );
        if( keyword == "section" ) {
            cfgerr = ( !eos || tokens.size() > 2 || !section.empty() );
            section = tokens[1];
            eos = false;
            continue;
        }
        if( keyword == "subsection" ) {
            cfgerr = ( !eoss || tokens.size() > 2 || !subsection.empty() );
            subsection = tokens[1];
            eoss = false;
            continue;
        }
        if( keyword == "endsection" ) {
            cfgerr = ( !eoss || tokens.size() > 1 || section.empty() );
            section.erase();
            eos = true;
            continue;
        }
        if( keyword == "endsubsection" ) {
            cfgerr = ( tokens.size() > 1 || subsection.empty() );
            subsection.erase();
            eoss = true;
            continue;
        }
        cfgerr = ( eos || eoss );
        cfgarr[section][subsection][tokens[0]] = tokens;
        std::ostringstream oss;
        oss << section << '/' << subsection << '/' << tokens[0] << " =";
        for(unsigned int i=0; i < tokens.size()-1; i++ ) oss << ' ' << tokens.at(i+1) ;
        pr.push_back(oss.str());
    }
    return( nstr );
}

Config::Config( const std::string& fname ) {
//    cerr << " Configs\n";
    Prow.reserve(100);
    cfg_file_read( fname, conf, Prow );
}

bool Config::issection( const std::string& sect ) {
    return conf.find( sect ) != conf.end(); 
}

bool Config::issubsection( const std::string& sect, const std::string& ssect ) {
//    Tree::const_iterator it = conf.find(sect);
//    if( it == conf.end() ) return false;
    return issection( sect ) && conf[sect].find(ssect) != conf[sect].end();
}

bool Config::isparameter( const std::string& sect, const std::string& ssect, const std::string& param) {
    return issubsection( sect, ssect ) && conf[sect][ssect].find(param) != conf[sect][ssect].end();
}

std::string Config::copy(const std::string& sect, const std::string& subs, const std::string& item ) {
    std::vector<std::string> *values = &conf[sect][subs][item];
    return values->at(1); 
}

Geo Config::site_coords() {
    std::vector<std::string> l = inpv<std::string>( "General", "Site", "Longitude" );
    std::string slon = l[0] + " " + l[1] + " " + l[2];
    l = inpv<std::string>( "General", "Site", "Latitude" );
    std::string slat = l[0] + " " + l[1] + " " + l[2];
    return Geo( Hour(slon), Degree(slat) );
}

int Tool::parse( const char *cs, vector<string>& item ) {
    string::size_type i = 0, j = 0;
    int k = 0;
    string s(cs);
    string u;
    while( (i =  s.find( "= ",i )) != string::npos ) s.replace(i,2,"=");
    i=0;
    while( (i = s.rfind( " =",i )) != string::npos ) s.replace(i,2,"=");
    i=0;
    while( (i = s.find( "\\\"",i )) != string::npos )  s.replace(i,2,"''");
    j=0;
    item.clear();
    while( (i = s.find_first_not_of( " \t,",j )) != string::npos ) {
        if( s.at(i)==';' || s.at(i)=='#') break;
        item.resize(k+1);
        if( (j = s.find_first_of( " \t,\"",i )) != string::npos && s.at(j) =='\"'){
            s.erase(j,1);
            if( (j = s.find_first_of( "\"",j )) != string::npos ) s.erase(j,1);
            else return( 0 );
            u = s.substr(i,j-i);
        } else {
            if( s.find_first_of( "#;",i ) < j ) j = s.find_first_of( "#;",i );
            u = s.substr(i,j-i);
        }
        if( u == "yes" || u == "on" ) u = "true";
        if( u == "no" || u == "off" ) u = "false";
        item[k++] = u;
    }
    return k;
}

int Tool::item_cnt( const char *cs ) {
    string::size_type i = 0, j = 0;
    int k = 0;
    string s(cs);
    string u;
    while( (i =  s.find( "= ",i )) != string::npos ) s.replace(i,2,"=");
    i=0;
    while( (i = s.rfind( " =",i )) != string::npos ) s.replace(i,2,"=");
    i=0;
    while( (i = s.find( "\\\"",i )) != string::npos )  s.replace(i,2,"''");
    j=0;
    while( (i = s.find_first_not_of( " \t,",j )) != string::npos ) {
        if( s.at(i)==';' || s.at(i)=='#') break;
        if( (j = s.find_first_of( " \t,\"",i )) != string::npos && s.at(j) =='\"'){
            s.erase(j,1);
            if( (j = s.find_first_of( "\"",j )) != string::npos ) s.erase(j,1);
            else return( 0 );
        } else {
            if( s.find_first_of( "#;",i ) < j ) j = s.find_first_of( "#;",i );
        }
        k++;
    }
    return k;
}

#include <regex.h>

void Tool::decompress( std::string &s ) {
    regex_t rr;
    int ret=0;
    size_t n = 4;
    regmatch_t regm[4];

    if( ( ret = regcomp( &rr, "([0-9]+)\\*(\\(([a-z+]+)\\)|[a-z])", REG_ICASE | REG_EXTENDED )) != 0 ){
        throw ErrFat( ERR_REG_STRUCT, "([0-9]+)\\*(\\(([a-z+]+)\\)|[a-z])" );
    }

    while( ( ret = regexec( &rr, s.c_str(), n, regm, 0)) == 0 ){
        regmatch_t rm = ( regm[3].rm_so >= 0 ? regm[3] : regm[2] );
        std::string tmp( s.substr(0,regm[0].rm_so) );
        tmp += s.substr(rm.rm_so,rm.rm_eo-rm.rm_so);
        for( int i=0; i < atoi(s.substr(regm[1].rm_so,regm[1].rm_eo-regm[1].rm_so).c_str())-1;i++){
            tmp += "+" +s.substr(rm.rm_so,rm.rm_eo-rm.rm_so);
        }
        tmp += s.substr(regm[0].rm_eo,s.length());
        s = tmp;
    }
    std::string::size_type i = 0;
    while( (i = s.find_first_of("()+",i)) != std::string::npos ) s.erase(i,1);
}




