#include <iostream>
#include <fstream>

#include "coordinates.h"

AstroDateTime Geo::utc;

double airmass( double ha, double dec, double lat ) {
    double secz1 = 1.0/(sin(lat)*sin(dec) + cos(lat)*cos(dec)*cos(ha));
    double secz2 = secz1 - 1.0;
    return secz1 - secz2*( 0.0018167 + secz2*( 0.002875 + 0.0008083*secz2 ) );
}

namespace std{
    static const int format_index = std::ios_base::xalloc();
    static const int separ_index  = std::ios_base::xalloc();

    std::ostream& operator<<(std::ostream& stm, const Hour& hh ){
                        // mapping to interval 0-2*Pi: (M_PI)*(1-sign(a))+fmod(a,(2*M_PI))
        double x = Hour::R*nint(hh);
        char sep = stm.iword(separ_index)? ':' : ' ';
        stm << std::setw(2) << std::setfill('0') << (int)x << sep << std::setw(2) << (int)fmod( 60*x, 60 )
            << sep << std::setw(2) << (int)fmod( fmod( 3600*x, 3600 ), 60 ) << std::setfill(' ');
        return stm;
    }
    
    std::ostream& operator<<(std::ostream& stm, const Degree& dd ){
                        // mapping to interval +/-Pi: sign(a)*(fmod(fabs(a)+M_PI,(2*M_PI))-M_PI);
        double x = stm.iword(format_index)?  Degree::R*nint(dd) : Degree::R*sint(dd);
        if( stm.iword(format_index) ) { stm << std::setw(3); stm.iword(format_index) = false; }
        else stm << std::setw(0) << ((x > 0) ? '+' : '-' ) << std::setw(2);
        stm << std::setfill('0') << (int)fabs(x) << ' ' << std::setw(2) << (int)fmod( 60*fabs(x), 60 )
            << ' ' << std::setw(2) << (int)fmod( fmod( 3600*fabs(x), 3600 ), 60 ) << std::setfill(' ');
        return stm;
    }
    
    std::istream& operator>> ( std::istream& stm, Hour& hh ){
        std::istream::sentry se(stm);
        if( se ){
            bool sign = false;
            if( stm.peek() == '-' ) {
                sign = true;
                stm.ignore();
            }
            unsigned int hour, min, sec;
            stm >> hour >> min >> sec;
        //FIXME:
            hh = ( sign ? -1.0 : 1.0 ) / ((12.0*3600)/M_PI) * ( hour*3600 + min*60 + sec );
        }
        return stm;
    }

    std::istream& operator>> ( std::istream& stm, Degree& dd ){
        std::istream::sentry se(stm);
        if( se ){
            bool sign = false;
            if( stm.peek() == '-' ) {
                sign = true;
                stm.ignore();
            }
            unsigned int deg, min, sec;
            stm >> deg >> min >> sec;
        //FIXME:
            dd = ( sign ? -1.0 : 1.0 ) / ((180.0*3600)/M_PI) * ( deg*3600 + min*60 + sec );
        }
        return stm;
    }
    
    std::istream& operator>>(std::istream& stm, Date& dd ){
        std::string scopy;
        stm >> scopy;
        std::replace( scopy.begin(), scopy.end(), '-', ' ' );
        std::istringstream strm(scopy);
        strm >> dd.year >> dd.month >> dd.day;
        return stm;
    }
    
    std::istream& operator>>( std::istream& stm, Time& tt ){
        std::string scopy;
        stm >> scopy;
        std::replace( scopy.begin(), scopy.end(), ':', ' ' );
        std::istringstream strm(scopy);
        strm >> tt.hour >> tt.min >> tt.sec;
        return stm;
    }
    
    std::ostream& operator<<(std::ostream& stm, const Date& dd ){
        stm << std::setfill('0') << std::setw(4) << dd.year  << '-' << std::setw(2)
            << dd.month << '-' << std::setw(2) << dd.day << std::setfill(' ');
        return stm;
    }
    
    std::ostream& operator<<(std::ostream& stm, const Time& tt ){
        stm << std::setfill('0') << std::setw(2) << tt.hour << ':' << std::setw(2)
            << tt.min << ':' << std::setw(2) << tt.sec << std::setfill(' ');
        return stm;
    }
    
    std::ostream& sdeg( std::ostream& stm ){
        stm.iword(format_index) = false;
        return stm;
    }
    
    std::ostream& udeg( std::ostream& stm ){
        stm.iword(format_index) = true;
        return stm;
    }
    
    std::ostream& colon( std::ostream& stm ){
        stm.iword(separ_index) = true;
        return stm;
    }
};

