#ifndef _MODULE_H
#define _MODULE_H

#include <configs.h>
#include <rs485.h>

class Module {
    static std::map< int, std::string > names; ///<       .
    static const SComm RESET;
    static const AByte GET_STATUS;
    static const AByte GET_SENSOR;
    static const AData GET_IDENT;
    static const AData GET_CONST;
    RS485& rs;
    std::string mod_name;
    unsigned int  mod_address;
    bool mod_verified; ///<    
    bool mod_connected;
    std::vector<unsigned char> ident; ///< 4    
protected:
    int status; ///<     
    int sensor; ///<     /
    static bool parking;
    std::vector<unsigned char> constants;
    static const std::string sect_device;
    void readConst();
    static void nsleep(double);
private:
    Module(const Module&);
    void signal_handler( RS485::ErrSignal& e );
    pthread_t Conn_thread;
    static void* thread( void* );
    void verification();
public:
    class Lock { // for parking
        Lock(const Lock&);
        Lock& operator=(const Lock&);
        static pthread_mutex_t mtx;
        struct timespec locktime;
    public:
        Lock() {
            locktime.tv_sec = time(0)+5;
            locktime.tv_nsec = 0;
            int err;
            if((err=pthread_mutex_timedlock(&mtx, &locktime)) != 0) {
                if(err == ETIMEDOUT) ErrlogSingl::instance().write( "Module Lock 5 sec timeout" );
                else ErrlogSingl::instance().write( "Module Lock 5 sec unknown error" );
            }
        }
        ~Lock() {pthread_mutex_unlock(&mtx);}
    };
    Module( Config&, RS485& , const std::string&);
    virtual ~Module ();
    bool connection() { return mod_connected; }
    virtual void reset();
    static const std::string& name( int a );
    const std::string& name() const { return (mod_name); }
    unsigned int address() { return mod_address; }
protected:
    int askByte( AByte c ) try {
        return rs.askByte( mod_address, c );
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return askByte(c);
//        return -1;
    }

    int askWord( AWord c ) try {
        return rs.askWord( mod_address, c );
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return askWord(c);
//        return -1;
    }

    int askLong( ALong c ) try {
        return rs.askLong( mod_address, c );
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return askLong(c);
//        return -1;
    }

    std::vector<unsigned char> askData( AData c ) try {
        return rs.askData( mod_address, c );
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return askData(c);
//        return  std::vector<unsigned char>();
    }

    int sendSimpleCommand( SComm c ) try {
        return rs.sendSimpleCommand( mod_address, c);
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return sendSimpleCommand(c);
//        return -1;
    }

    int sendByteCommand( SByte c, char b) try {
        return rs.sendByteCommand( mod_address, c, b);
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return sendByteCommand(c,b);
//        return -1;
    }

    int sendWordCommand( SWord c, short w) try {
        return rs.sendWordCommand( mod_address, c, w);
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return sendWordCommand(c, w);
//        return -1;
    }

    int sendData( std::vector<unsigned char> &d ) try {
        return rs.sendData( mod_address, d.size(), d );
    }
    catch ( RS485::ErrSignal& e ) {
        signal_handler( e );
        return sendData(d);
//        return -1;
    }

    int getStatus() {
        return status = askByte(GET_STATUS);
    }

    int getSensor() {
        return sensor = askByte(GET_SENSOR);
    }
};

#endif

