#ifndef __SERVER_H_
#define __SERVER_H_

#include <string>
#include <semaphore.h>

#include <socket.h>

class ClientHandler;

typedef std::string(*CALLF_H)( Client*, std::string );

typedef void(ClientHandler::*CALLF_C)( std::string );

class Client;

class Server {
private:
    CBTableServSock cbtable_socket;
    CALLF_H comm_handl;
    ServerSocket* m_socket;
    std::list<Client*> m_clients;
    std::stack<Client*> m_clientsToRemove;
    Server( const Server& );
    Server& operator=( const Server& );
    bool _quit;
public:
    Server( const ServerSocket::AddrType&, const CALLF_H );
    ~Server();
    bool isQuit() { removeDisconnectedClients(); return _quit; } 
private:
    void removeDisconnectedClients();
    void listen( const ServerSocket::AddrType& );
    void close();
    ServerSocket* socket() { return m_socket; }
    void quit() { _quit = true; }
    void onSocketStateChanged( const SocketBase::SocketState s );
    void onSocketAccepted();
    void onSocketError( const SocketError& e );
friend class Client;
private:
    void onClientClosed( Client* c );
    void onClientRequest( Client* c );
};

class Client {
private:
    CBTableClientSock cbtable_socket;
    Server* m_parent;
    ClientSocket* m_socket;
    Client( const Client& );
    Client& operator=( const Client& );
    std::string buff_mesgs;
    ClientSocket::AddrType m_addr;
    CALLF_C comm_handl;
    ClientHandler* handclass;
    sem_t client_sem;
public:
    Client( Server* );
    Client( ClientSocket::AddrType&, ClientHandler* );
    ~Client();
    void sendResponse( const std::string& );
    void sendRequest( const std::string& );
    void sendRequest( const std::string&, const CALLF_C );
    void close();
    void quit() { m_parent->quit(); }
    void waitReply( unsigned int );
    ClientSocket* getSocket() {return m_socket ;}
    void storeAddr( ClientSocket::AddrType& a){ m_addr = a;}
    bool isConnected();
private:
    void onSocketStateChanged( const SocketBase::SocketState s );
    void onSocketDataRecieved();
    bool messExtract( std::string& );
    void onSocketError( const SocketError& e );
};

#endif


