#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <errlog.h>
#include <astrotime.h>
#include <configs.h>
#include <signal.h>

#include "sub_main.h"
#include "deftypes.h"
#include "defnames.h"
#include "processes.h"
#include "conditions.h"

namespace Visor {
    pthread_t Observ;
    void* thread( void* );
    bool  works = true;
    void catch_fault(int sig);
    bool expire = false;
    bool back_mode = false;
    static sigset_t signal_mask;
}

int Visor::cmd_line( int ac, char **av ) {
    opterr = 0;
    int c, errcount = 0;
    while((c = getopt(ac, av, "hbdmu:s:")) != -1)  {
        switch(c) {
            case 'h':   std::cerr << "Read manual, at least" << std::endl;
                        errcount++;
                        break;
            case 'b':   back_mode = true;
                        break;
            case 'd':   Observation::cmd.test_mode = true;
                        break;
            case 'm':   Observation::cmd.man_mode = true;
                        break;
            case 'u':   if( optarg != NULL ) {
                            Condition::cmd.final_time = time24( 12*Hour(std::string(optarg) )/M_PI );
//                            if( Condition::cmd.final_time < Condition::loc_time() ) {
                            if( Condition::cmd.final_time < Condition::utc() ) {
                                ErrlogSingl::instance().write( ERR_CMD_OPTION, "Wrong finish time" );
                                errcount++;
                            }
                        }
                        break;
            case 's':   if( optarg != NULL ) {
                            Condition::cmd.start_time = time24( 12*Hour( std::string(optarg) )/M_PI );
                        }
                        break;
            default:    ErrlogSingl::instance().write( ERR_CMD_OPTION, "Command line" );
                        errcount++;
                        break;
        }
    }
    ErrlogSingl::instance().debug(Observation::cmd.test_mode);
    if( setreuid(geteuid(), geteuid()) < 0 )
        { errcount++ ; ErrlogSingl::instance().write( (ErrorsCode)errno, "Can not set mass user " ); }
    if( setregid(getegid(), getegid()) < 0 )
        { errcount++ ; ErrlogSingl::instance().write( (ErrorsCode)errno, "Can not set mass group" ); }
    if( access(program_etc.c_str(), F_OK|R_OK ) < 0 )
        { errcount++ ; ErrlogSingl::instance().write( (ErrorsCode)errno, std::string(program_etc) ); }
    if( access(program_log.c_str(), F_OK|W_OK ) < 0 )
        { errcount++ ; ErrlogSingl::instance().write( (ErrorsCode)errno, std::string(program_log) ); }
    if( access( Observation::cmd.cfgfile.c_str(), F_OK|R_OK ) < 0 )
        { errcount++ ; ErrlogSingl::instance().write( (ErrorsCode)errno, Observation::cmd.cfgfile ); }
    if( access( Condition::cmd.cfgfile.c_str(), F_OK|R_OK ) < 0 )
        { errcount++ ; ErrlogSingl::instance().write( (ErrorsCode)errno, Condition::cmd.cfgfile ); }
    if( back_mode && errcount == 0 ) {
        ErrlogSingl::instance().write( "Program demonization ... done" );
        int pid = fork();
        if (pid < 0) {
            ErrlogSingl::instance().write( (ErrorsCode)errno, "Call fork() failed " );
            exit(1);
        }
        if (pid == 0) {
            setsid();
            close( 0 ); close( 1 ); close( 2 );   // redirect stdout/stderr etc to /dev/null
            int fd = open("/dev/null", O_RDWR);
            dup( fd ); dup( fd ); dup( fd );
            errcount = 0;
        } else {
            errcount = 1;
        }
    }
    return errcount;
}

void Visor::create() {
    sigemptyset( &signal_mask );
    sigaddset( &signal_mask, SIGTERM );
    sigaddset( &signal_mask, SIGHUP );
    pthread_sigmask( SIG_BLOCK, &signal_mask, NULL);
    pthread_create( &Observ, NULL, Visor::thread,  NULL );
}

bool Visor::isQuit(){
    return !works;
}

void Visor::catch_fault(int sig) {
    if( sig == SIGTERM ) {
        expire = true;
        Observation::terminate = true;
        Condition::terminate = true;
        ErrlogSingl::instance().write( "SIGTERM catched! Observation will be finished" );
    }
    if( sig == SIGHUP ) {
        ErrlogSingl::instance().reopen();
        ErrlogSingl::instance().write( "SIGHUP catched! New logfile was created if date changes" );
    }
}

void *Visor::thread( void* a ) {
    works = true;
    pthread_detach( Observ );
    signal( SIGTERM, catch_fault );
    signal( SIGHUP,  catch_fault );
    pthread_sigmask (SIG_UNBLOCK, &signal_mask, NULL);
    try {
        Condition condition;
        message<Hour,0>( "Start time: ",  (Hour)(M_PI*condition.cmd.start_time/12.0),
                         "Finish time: ", (Hour)(M_PI*condition.cmd.final_time/12.0) );
        while( !expire && Condition::utc() < condition.cmd.watch_time ) {
            condition.wait_conditions();
            ErrlogSingl::instance().write( "Conditions are good, ASM start" );
            try {
                Observation observation;
                while( !expire && condition.evening_time() && condition.good() ) observation.experiment();
                while( !expire && condition.night_time()   && condition.good() ) {
                    try {
                        observation.programm();
                    }
                    catch (const std::exception& e ) {
                        ErrlogSingl::instance().write( e.what() );
                        message<Hour,0>( "ASM pauses until: ", (Hour)(M_PI*condition.restart_time()/12.0) );
                        condition.wait_conditions();
                    }
                }
                while( !expire && condition.morning_time() && condition.good() ) observation.experiment();
                                    // finish contitions 1) signal 2) time pass away 3) bad wether 4) bad data
                if( !expire && condition.bad() ) ErrlogSingl::instance().write( "Conditions are bad. Wait" );
            }
            catch (const std::exception& e ) {
                ErrlogSingl::instance().write( e.what() );
                message<Hour,0>( "ASM waits until: ", (Hour)(M_PI*condition.restart_time()/12.0) );
            }
        }
    }
    catch (const std::exception& e ) {
        ErrlogSingl::instance().write( e.what() );
    }
    sleep(4);
    works = false;
    return 0;
}
